// Permutations GUI
// by Jason Riggs, Jacob Quinn Shenker, Jay Bhat
//
// For a full description of the project, visit:
// http://cm-wiki.stanford.edu/wiki/Permutations
//

// create our OSC receiver
OscRecv recv;
6449 => recv.port;
recv.listen();
recv.event( "/sync/clock, i, f" ) @=> OscEvent oe;

0 => int play;
int beat;
dur T_server;

fun void serverListener() 
{
    while ( true )
    {
        oe => now;
        while( oe.nextMsg() != 0 )
        {
            oe.getInt() => beat;
			oe.getFloat()::second => T_server;
            doBeat();
        }
    }
}
spork ~ serverListener();

// OSC

OscSend send;
send.setHost("localhost", 7007);
//send.event( "/sync/clock, i i" ) @=> OscEvent clock_event;
//send.event( "/sync/beat, i i" ) @=> OscEvent beat_event;
//send.event( "/inst/filter, i" ) @=> OscEvent filter_event;
//send.event( "/inst/detune, i" ) @=> OscEvent detune_event;
//send.event( "/inst/play, f f i f f f f" ) @=> OscEvent play_event;

// basenote range
60.0 + 6*12 => float kMaxBasenote;
60.0 - 4*12 => float kMinBasenote;
60.0 => float kDefaultBasenote;

// default envelope value
2 => int kDefaultEnvelope;
0 => int kDefaultTimbre;
4 => int kMaxTimbres;

// timbre
kDefaultTimbre => int timbre;

[ [2.0, 1.6, .75, 2.0],
[.001, 0.01, 0.0, 0.0],  
[.001, .1, 0.0, 0.0], 
[.001, .2, 0.0, 0.0], 
[.001, .4, 0.0, 0.0], 
[.1, .4, 0.0, 0.0], 
[.25, .1, .95, .25], 
[.5, .2, .9, .5], 
[1.0, .4, .85, 1.0], 
[1.5, .8, .8, 1.5] ] @=> float envelopes[][];

envelopes[kDefaultEnvelope][0]::second => dur att;
envelopes[kDefaultEnvelope][1]::second => dur dec;
envelopes[kDefaultEnvelope][2] => float sus;
envelopes[kDefaultEnvelope][3]::second => dur rel;

// keyboard modifier keys
0 => int button_action;

// tempo for autoplay
32.0 => float T;

// note arrays (rows of GUI)
//[[0, 2, 3, 7, 8, 13], 
//[0,1,2,3],
//[0,-12,0,12]] @=> int rows[][];
// experimentation below
[[16, 12, 9, 5],
[0, -12, 0, -12],
[0, -12, 0, 12]] @=> int rows[][];
[0, 5, 0] @=> int row_format[];

// positions
int positions[rows.cap()];

// enable trackpad input (initially disabled)
0 => int trackpadEnabled;

// basenote stuff
kDefaultBasenote => float basenote;

// sensitivity (instrument control)
.008 => float detuneSensitivity;
24.0 => float lpfSensitivity;

// keyboard thing
1 => int kKbdSpacebar;

// tempo range allowed (expressed as time between notes)
1000 => int kMaxTempo;
1 => int kMinTempo;

MAUI_View control_view;
MAUI_LED led_rows[rows.cap()][0];//[rows[0].cap()];
MAUI_Button button_rows[rows.cap()][0];//[rows[0].cap()];
control_view.size(rows[0].cap()*100, rows.cap()*100+50+50+50);
MAUI_Button dummy_button;
dummy_button.size(1,1);
dummy_button.position(0,0);
control_view.addElement(dummy_button);

MAUI_Button play_button;
play_button.toggleType();
play_button.name("play (tilde)");
play_button.size(control_view.width()/3., 100);
play_button.position(0, control_view.height()-100);
control_view.addElement(play_button);
fun void controlPlayButton()
{
    while (true)
    {
        play_button => now;
        <<< "play button:", play_button.state() >>>;
        togglePlay(play_button.state());
    }
}
spork ~ controlPlayButton();
MAUI_Button spacebar_button;
spacebar_button.pushType();
spacebar_button.name("next (space)");
spacebar_button.size(control_view.width()/3., 100);
spacebar_button.position(control_view.width()/3., control_view.height()-100);
control_view.addElement(spacebar_button);
fun void controlSpacebarButton()
{
    while (true)
    {
        spacebar_button => now;
        doBeat();
    }
}
spork ~ controlSpacebarButton();

for (0 => int i; i < rows.cap(); i++)
{
    MAUI_LED leds[rows[i].cap()] @=> led_rows[i];
    MAUI_Button buttons[rows[i].cap()] @=> button_rows[i];
    for (0 => int k; k < rows[i].cap(); k++)
    {
        leds[k] @=> MAUI_LED led;
        led.color(MAUI_LED.blue);
        led.size(30,30);
        led.position(k*100+20,i*100+35);
        control_view.addElement(led);
        buttons[k] @=> MAUI_Button button;
        button.pushType();
        button.size(100,100);
        button.position(k*100,i*100+50);
        //updateButton(i, k);
        control_view.addElement(button);
    }
}
updateDisplay();
control_view.display();

fun void change_action(int increment)
{
    increment +=> button_action;
    if (button_action < 0)
    {
        0 => button_action;
    }
}
fun void togglePlay(int state)
{
    state => play;
}


fun void updateButton(int i, int k)
{
    button_rows[i][k] @=> MAUI_Button button;
    button.name(format_note(i, k));
}

fun int get_absolute_pitch(int i, int k)
{
	0 => int pitch;
	for (rows.cap() - 1 => int j; j > i; j--)
	{
		rows[j][positions[j]] +=> pitch;
	}
	rows[i][k] +=> pitch;
	return pitch + (Math.trunc(basenote) $ int) - 60; // pitch relative to middle C
}

fun int get_pitch_class(int pitch)
{
	while (pitch < 0) { 12 +=> pitch; }
	return pitch % 12;
}

fun string repeat_string(string str, int times)
{
	string out;
	for (0 => int i; i < times; i++)
	{
		str +=> out;
	}
	return out;
}

fun string format_octave(int pitch)
{
	"" => string str;
	Math.ceil((pitch - 11) / 12.0) $ int => int octaves;
	if (octaves > 0)
	{
		repeat_string("'", octaves) +=> str;
	}
	else if (octaves < 0)
	{
		repeat_string(",", -octaves) +=> str;
	}
	return str;
}

fun string format_pitch_class(int pitch)
{
	// this can't be declared globally, or ChucK crashes
	//["c","des","d","ees","e","f","ges","g","aes","a","bes","b"] @=> string note_names[];
	["c","cis","d","dis","e","f","fis","g","gis","a","ais","b"] @=> string note_names[];
	return note_names[get_pitch_class(pitch)];
}

fun string format_interval(int pitch)
{
	Math.ceil((pitch - 11)/ 12.0) $ int => int octaves;
	get_pitch_class(pitch) => int pitch_class;
	// format M3 + 8va -15ma, etc. (23vb??)
	string quality;
	int interval;
	//["P1","m2","M2","m3","M3","P4","TT","P5","m6","M6","m7","M7","P8"] @=> string note_names[]; // this can't be declared globally, or ChucK crashes
	if (pitch_class % 12 == 0)
	{ "P" => quality; 1 => interval; }
	else if (pitch_class == 1) { "m" => quality; 2 => interval; }
	else if (pitch_class == 2) { "M" => quality; 2 => interval; }
	else if (pitch_class == 3) { "m" => quality; 3 => interval; }
	else if (pitch_class == 4) { "M" => quality; 3 => interval; }
	else if (pitch_class == 5) { "P" => quality; 4 => interval; }
	else if (pitch_class == 6) { "TT" => quality; 0 => interval; }
	else if (pitch_class == 7) { "P" => quality; 5 => interval; }
	else if (pitch_class == 8) { "m" => quality; 6 => interval; }
	else if (pitch_class == 9) { "M" => quality; 6 => interval; }
	else if (pitch_class == 10) { "m" => quality; 7 => interval; }
	else if (pitch_class == 11) { "M" => quality; 7 => interval; }
	string str;
	if (octaves > 0) { repeat_string("'", octaves) +=> str; }
	else if (octaves < 0) { repeat_string(",", -octaves) +=> str; }
	quality +=> str;
	if (interval != 0) { interval + "" +=> str; }
	return str;
}

fun string format_note(int i, int k)
{
	if (row_format[i] == 0) // absolute pitches
	{
		return format_octave(get_absolute_pitch(i, k)) + format_pitch_class(get_absolute_pitch(i, k));
	}
	else if (row_format[i] == 1) // absolute pitch classes (no octave information)
	{
		return format_pitch_class(get_absolute_pitch(i, k));
	}
	else if (row_format[i] == 3) // absolute intervals from middle C
	{
		return format_interval(get_absolute_pitch(i, k));
	}
	else if (row_format[i] == 4) // absolute intervals from middle C as integer (unformatted)
	{
		return get_absolute_pitch(i, k) + "";
	}
	else if (row_format[i] == 5) // relative intervals
	{
		return format_interval(rows[i][k]);
	}
	else if (row_format[i] == 6) // relative intervals as integer (unformatted; raw value in rows[i][k])
	{
		return rows[i][k] + "";
	}
}

fun void controlButton(int i, int k)
{
    button_rows[i][k] @=> MAUI_Button button;
    while( true )
    {
        // wait for the button to be pushed down
        button => now; // button down
        <<< "button pushed:", button.name(), button_action >>>;
        if (button_action == 0) // click: swap k-th and (k+1)-th (modulo length of tone row)
        {
            (k+1) % rows[i].cap() => int k1;
            rows[i][k] @=> int temp;
            rows[i][k1] @=> rows[i][k];
            temp @=> rows[i][k1];
            //updateButton(i, k1);
        }
        else if (button_action == 4) // click+shift: increment
        {
            1 +=> rows[i][k];
        }
        else if (button_action == 8) // click+apple: decrement
        {
            1 -=> rows[i][k];
        }
        else if (button_action == 6) // click+opt+shift: increment by an octave
        {
            12 +=> rows[i][k];
        }
        else if (button_action == 10) // click+opt+apple: decrement by an octave
        {
            12 -=> rows[i][k];
        }
		else if (button_action == 5) // click+ctl+shift: increment by a fifth
        {
            7 +=> rows[i][k];
        }
		else if (button_action == 9) // click+ctl+apple: decrement by a fifth
        {
            7 -=> rows[i][k];
        }
		else if (button_action == 7) // click+ctl+opt+shift: increment by a major third
        {
            4 +=> rows[i][k];
        }
		else if (button_action == 11) // click+ctl+opt+apple: decrement by a major third
        {
            4 -=> rows[i][k];
        }
        //updateButton(i, k);
		// TODO: only rename what is necessary
		updateButtons();
        button => now; // button up
    }
}

for (0 => int i; i < rows.cap(); i++)
{
    for (0 => int k; k < rows[i].cap(); k++)
    {
        spork ~ controlButton(i, k);
    }
}

fun void updatePosition(MAUI_LED leds[], int position)
{
    for (0 => int i; i < leds.cap(); i++)
    {
        if (position == i)
        {
            leds[i].light();
        }
        else
        {
            leds[i].unlight();
        }
    }
}

fun void updateDisplay()
{
    updatePositions();
	updateButtons();
}

fun void updatePositions()
{
	for (0 => int i; i < rows.cap(); i++)
    {
        updatePosition(led_rows[i], positions[i]);
    }
}

fun void updateButtons()
{
	for (0 => int i; i < rows.cap(); i++)
    {
		for (0 => int k; k < rows[i].cap(); k++)
		{
			updateButton(i, k);
		}
    }
}

fun void advanceTime()
{
    1 => int advance;
    for (0 => int i; i < rows.cap(); i++)
    {
        if (advance)
        {
            (1 + positions[i]) % rows[i].cap() => positions[i];
        }
        if (positions[i] == 0 && advance)
        {
            1 => advance;
        }
        else
        {
            0 => advance;
        }
        
    }
}

fun void sendNote()
{
    0 => int note;
    for (0 => int i; i < rows.cap(); i++)
    {
        rows[i][positions[i]] +=> note;
    }
    oscSendPlay(Std.mtof(note+basenote), 0.0, timbre, att/1::second, dec/1::second, sus, rel/1::second);
}

0 => int interpolating;
1 => float display_limit; // i.e., no display limit

fun void doBeat()
{
	if (T < 1.0)
	{
		if (!interpolating) spork ~ interpolateTime();
	}
	else
	{
    	if (beat % (T $ int) == 0) 
    	{
       		advanceTime();
			if (T >= display_limit)
			{
        		updateDisplay();
			}
        	if(play) spork ~ sendNote();
    	}
	}
}

0 => int interpolated_beat;
64 => int interpolated_steps;
4 => int interpolated_display_limit;
dur T_interpolated;

fun void interpolateTime()
{
	0 => interpolated_beat;
	T_server/interpolated_steps => T_interpolated;
	1 => interpolating;
	now => time start_time;
	while (true)
	{
		if (T >= 1) break;
		if (T < 1.0/interpolated_steps)
		{
			//<<< "Tempo is below interpolation tempo; not interpolating" >>>;
			break;
		}
		if (interpolated_beat == interpolated_steps)
		{
			//<<< "resetting interpolation" >>>;
			interpolated_steps*T_interpolated +=> start_time;
			0 => interpolated_beat;
		}
		if ((interpolated_beat % (T*interpolated_steps $ int)) == 0)
		{
			//<<< "interpolated BEAT!!!", interpolated_beat >>>;
			spork ~ advanceTime();
			if (T >= 1.0/interpolated_display_limit)
			{
				spork ~ updateDisplay();
			}
			if (play) spork ~ sendNote();
		}
		//doInterpolatedBeat();
		//<<< start_time, interpolated_beat*T_interpolated, now, interpolated_beat*T_interpolated + start_time, ((start_time + interpolated_beat*T_interpolated) - now)/1.0::second, T_interpolated/1.0::second >>>;
		//((start_time + interpolated_beat*T_interpolated) - now) => now;
		T_interpolated => now;
		//0.5::second => now;
		interpolated_beat++;
		//<<< "interpolated hit", interpolated_beat >>>;
	}
	0 => interpolating;
}

spork ~ keyboardControl();
spork ~ trackpadControl();

// the device number to open
0 => int keyboardNum;

// instantiate a HidIn object
HidIn keyboardIn;
// structure to hold HID messages
HidMsg keyboardMsg;

// open keyboard
if( !keyboardIn.openKeyboard( keyboardNum ) ) me.exit();
// successful! print name of device
<<< "keyboard '", keyboardIn.name(), "' ready" >>>;

fun void keyboardControl() 
{
    while( true )
    {
        // wait on event
        keyboardIn => now;
        
        // get one or more messages
        while( keyboardIn.recv( keyboardMsg ) )
        {
            
            // if out of bounds, ignore command
            if( keyboardMsg.which > 256 ) 
            {
                continue;
            }
                        
            if (keyboardMsg.which == 53 && keyboardMsg.isButtonDown())
            {
                <<< "tilde down" >>>;
                play_button.state(!play_button.state());
                togglePlay(play_button.state());
            }
            else if (keyboardMsg.which == 44 && keyboardMsg.isButtonDown())
            {
                <<< "space down" >>>;
                if (kKbdSpacebar == 1)
                {
                    doBeat();
                }
            }
            else if (keyboardMsg.which == 44 && !keyboardMsg.isButtonDown())
            {
                <<< "space up" >>>;
                // do stuff
            }
            else if (keyboardMsg.which == 224 && keyboardMsg.isButtonDown())
            {
                <<< "ctl down" >>>;
                change_action(1);
            }
            else if (keyboardMsg.which == 224 && !keyboardMsg.isButtonDown())
            {
                <<< "ctl up" >>>;
                change_action(-1);
            }
            else if (keyboardMsg.which == 226 && keyboardMsg.isButtonDown())
            {
                <<< "opt down" >>>;
                change_action(2);
            }
            else if (keyboardMsg.which == 226 && !keyboardMsg.isButtonDown())
            {
                <<< "opt up" >>>;
                change_action(-2);
            }
            else if (keyboardMsg.which == 225 && keyboardMsg.isButtonDown())
            {
                <<< "shift down" >>>;
                change_action(4);
            }
            else if (keyboardMsg.which == 225 && !keyboardMsg.isButtonDown())
            {
                <<< "shift up" >>>;
                change_action(-4);
            }
            else if (keyboardMsg.which == 227 && keyboardMsg.isButtonDown())
            {
                <<< "apple down" >>>;
                change_action(8);
            }
            else if (keyboardMsg.which == 227 && !keyboardMsg.isButtonDown())
            {
                <<< "apple up" >>>;
                change_action(-8);
            }   
            
            // 'up' and 'down' control timbre, 'esc' resets
            else if (keyboardMsg.which == 41 && keyboardMsg.isButtonDown())
            {
                resetTempo();
            }
            else if( keyboardMsg.which == 81 || keyboardMsg.which == 82 )
            {
                if( keyboardMsg.which == 81 && keyboardMsg.isButtonDown() )
                {
                    if (button_action == 0) // arrow: multiply tempo by 2
                    {
                        changeTempo(2);
                    }
				    else if (button_action == 4) // arrow+shift: multiply tempo by 3
                    {
                        changeTempo(3);
                    }
                    else if (button_action == 8) // arrow+apple: multiply tempo by 5
                    {
                        changeTempo(5);
                    }
                    else if (button_action == 12) // arrow+shift+apple: multiply tempo by 7
                    {
                        changeTempo(7);
                    }
                }
                else if( keyboardMsg.which == 82 && keyboardMsg.isButtonDown() )
                {
                    if (button_action == 0) // arrow: divide tempo by 2
                    {
                        changeTempo(1.0/2);
                    }
				    else if (button_action == 4) // arrow+shift: divide tempo by 3
                    {
                        changeTempo(1.0/3);
                    }
                    else if (button_action == 8) // arrow+apple: dividy tempo by 5
                    {
                        changeTempo(1.0/5);
                    }
                    else if (button_action == 12) // arrow+shift+apple: divide tempo by 7
                    {
                        changeTempo(1.0/7);
                    }
                }
            }
            
            // select envelope
            else if( keyboardMsg.which >= 30 && keyboardMsg.which <= 39 && keyboardMsg.isButtonDown() )
            {
                setEnvelope(keyboardMsg.which);
            }
            
            // enbable/disable trackpad input
            else if( keyboardMsg.which == 55 )
            {
                if( keyboardMsg.isButtonDown() )
                {
                    1 => trackpadEnabled;
                    <<< "trackpadenabled" >>>;
                }
                else if( keyboardMsg.isButtonUp() )
                {
                    0 => trackpadEnabled;
                    <<< "trackpadDisabled" >>>;
                }
            }
            
            // timbre control
            else if( keyboardMsg.which == 47 || keyboardMsg.which == 48 )
            {
                if( keyboardMsg.which == 47 && keyboardMsg.isButtonDown() )
                {
                    changeTimbre(keyboardMsg.which);
                }
                else if( keyboardMsg.which == 48 && keyboardMsg.isButtonDown() )
                {
                    changeTimbre(keyboardMsg.which);
                }
            }
            
            // change basenote
			// key[keyboardMsg.which] == 0 && 
            else if( keyboardMsg.which != 29 )
            {
                if( keyboardMsg.which == 79 && keyboardMsg.isButtonDown() )
                {
                    if (button_action == 0) // arrow: increment basenote by an octave
                    {
                        changeBasenote(12);
                    }
				    else if (button_action == 4) // arrow+shift: increment basenote by 100 cents
                    {
                        changeBasenote(1);
                    }
                    else if (button_action == 8) // arrow+apple: increment basenote by 50 cents
                    {
                        changeBasenote(0.5);
                    }
                    else if (button_action == 12) // arrow+shift+apple: increment basenote by 25 cents
                    {
                        changeBasenote(0.25);
                    }
                }
                else if( keyboardMsg.which == 80 && keyboardMsg.isButtonDown() )
                {
                    if (button_action == 0) // arrow: decrement basenote by an octave
                    {
                        changeBasenote(-12);
                    }
				    else if (button_action == 4) // arrow+shift: decrement basenote by 100 cents
                    {
                        changeBasenote(-1);
                    }
                    else if (button_action == 8) // arrow+apple: decrement basenote by 50 cents
                    {
                        changeBasenote(-0.5);
                    }
                    else if (button_action == 12) // arrow+shift+apple: decrement basenote by 25 cents
                    {
                        changeBasenote(-0.25);
                    }
                }
            }
            
        }
    }
}

// -------------------end keyboard input-----------------------|

// -------------------trackpad input-----------------------|

// the device number to open
0 => int trackpadNum;

// instantiate a HidIn object
HidIn trackpadIn;
// structure to hold HID messages
HidMsg trackpadMsg;

// open mouse 0, exit on fail
if( !trackpadIn.openMouse( trackpadNum ) ) me.exit();
// successful! print name of device
<<< "mouse '", trackpadIn.name(), "' ready" >>>;

fun void trackpadControl()
{
    // infinite event loop
    while( true )
    {
        // wait on HidIn as event
        trackpadIn => now;
        
        // messages received
        while( trackpadIn.recv( trackpadMsg ) )
        {
            // mouse motion
            if( trackpadMsg.isMouseMotion() )
            {
                // axis of motion
                if( trackpadMsg.deltaX )
                {
                    //<<< "mouse motion:", trackpadMsg.deltaX, "on x-axis" >>>;
                    //filterSweep(trackpadMsg.deltaX);
					if (trackpadEnabled)
					{
						//<<< "filter:", trackpadMsg.deltaX >>>;
						oscSendFilter(trackpadMsg.deltaX);
					}
                }
                else if( trackpadMsg.deltaY )
                {
                    //<<< "mouse motion:", trackpadMsg.deltaY, "on y-axis" >>>;
                    //detuneSpreader(trackpadMsg.deltaY);
					if (trackpadEnabled)
					{
						//<<< "detune:", trackpadMsg.deltaY >>>;
						oscSendDetune(trackpadMsg.deltaY);
					}
                }
            }
            
            // mouse button down
            else if( trackpadMsg.isButtonDown() )
            {
                //<<< "mouse button", trackpadMsg.which, "down" >>>;
            }
            
            // mouse button up
            else if( trackpadMsg.isButtonUp() )
            {
                //<<< "mouse button", trackpadMsg.which, "up" >>>;
            }
            
            // mouse wheel motion (requires chuck 1.2.0.8 or higher)
            else if( trackpadMsg.isWheelMotion() )
            {
                // axis of motion
                if( trackpadMsg.deltaX )
                {
                    //<<< "mouse wheel:", trackpadMsg.deltaX, "on x-axis" >>>;
                }            
                else if( trackpadMsg.deltaY )
                {
                    //<<< "mouse wheel:", trackpadMsg.deltaY, "on y-axis" >>>;
                }
            }
        }
    }
}

// increment basenote (left arrow key)
fun void changeBasenote(float increment)
{
    if( ((basenote+increment) <= kMaxBasenote) && ((basenote+increment) >= kMinBasenote )) { increment +=> basenote; }
    updateButtons();
    <<< "Basenote: ", basenote >>>;
    //<<< "register:", register, "scale degree":, hh >>>;
}

// tempo (up and down arrow keys)
fun void changeTempo(float multiplier)
{
    if (multiplier > 1.0 || T >= 1.0/interpolated_steps) { multiplier *=> T; }
	if (multiplier < 1.0 && T < 1.0)
	{
		if (!interpolating) spork ~ interpolateTime();
	}
    <<< "Tempo: ", T >>>;
}

// reset tempo to nearest multiple of two
fun void resetTempo()
{
    Math.pow(2, Math.floor(Math.log2(T))) => float T_low;
    Math.pow(2, Math.ceil(Math.log2(T))) => float T_high;
    if (T - T_low > T_high - T) { T_high => T; }
    else { T_low => T; }
    <<< "Tempo: ", T >>>;
}

// timbre ("[" and ]")
fun void changeTimbre(int message)
{
    // "["
    if(message == 47)
    {
        // decrease timbre
        if(timbre > 0) timbre--;
        <<< "Current Timbre: ", timbre >>>;
    }
    
    // "]"
    if(message == 48)
    {
        // increase timbre
        if(timbre + 1 < kMaxTimbres) timbre++;
    	<<< "Current Timbre: ", timbre >>>;
    }
}

// envelopes (digit keys)
fun void setEnvelope(int message)
{
    // find digit
    int digit;
    if(message == 39) 0 => digit;
    else (message - 30 + 1) => digit;
    
    // set envelope
    envelopes[digit][0]::second => att;
    envelopes[digit][1]::second => dec;
    envelopes[digit][2] => sus;
    envelopes[digit][3]::second => rel;
    
    <<< "Enevelope set: ", digit >>>;
}

// OSC handlers

fun void oscSendFilter(int change)
{
	send.startMsg("/inst/filter, i");
	send.addInt(change);
}

fun void oscSendDetune(int change)
{
	send.startMsg("/inst/detune, i");
	send.addInt(change);
}

fun void oscSendPlay(float freq, float velocity, int timbre, float att, float dec, float sus, float rel)
{
	send.startMsg("/inst/play, f, f, i, f, f, f, f");
	send.addFloat(freq);
	send.addFloat(velocity);
	send.addInt(timbre);
	send.addFloat(att);
	send.addFloat(dec);
	send.addFloat(sus);
	send.addFloat(rel);
}

while(true)
{
	1::day => now;
}